import six

from .sorted_dict import SortedDict
from .generic_repr import GenericRepr


class BaseFormatter(object):
    def can_format(self, value):
        raise NotImplementedError()

    def format(self, value, indent, formatter):
        raise NotImplementedError()

    def get_imports(self):
        return ()

    def assert_value_matches_snapshot(self, test, test_value, snapshot_value):
        test.assert_equals(test_value, snapshot_value)

    def store(self, test, value):
        return value


class TypeFormatter(BaseFormatter):
    def __init__(self, types, format_func):
        self.types = types
        self.format_func = format_func

    def can_format(self, value):
        return isinstance(value, self.types)

    def format(self, value, indent, formatter):
        return self.format_func(value, indent, formatter)


def trepr(s):
    text = '\n'.join([repr(line).lstrip('u')[1:-1] for line in s.split('\n')])
    quotes, dquotes = "'''", '"""'
    if quotes in text:
        if dquotes in text:
            text = text.replace(quotes, "\\'\\'\\'")
        else:
            quotes = dquotes
    return "%s%s%s" % (quotes, text, quotes)


def format_none(value, indent, formatter):
    return 'None'


def format_str(value, indent, formatter):
    if '\n' in value:
        # Is a multiline string, so we use '''{}''' for the repr
        return trepr(value)

    # Snapshots are saved with `from __future__ import unicode_literals`,
    # so the `u'...'` repr is unnecessary, even on Python 2
    return repr(value).lstrip('u')


def format_std_type(value, indent, formatter):
    return repr(value)


def format_dict(value, indent, formatter):
    value = SortedDict(**value)
    items = [
        formatter.lfchar + formatter.htchar * (indent + 1) + formatter.format(key, indent) + ': ' +
        formatter.format(value[key], indent + 1)
        for key in value
    ]
    return '{%s}' % (','.join(items) + formatter.lfchar + formatter.htchar * indent)


def format_list(value, indent, formatter):
    items = [
        formatter.lfchar + formatter.htchar * (indent + 1) + formatter.format(item, indent + 1)
        for item in value
    ]
    return '[%s]' % (','.join(items) + formatter.lfchar + formatter.htchar * indent)


def format_tuple(value, indent, formatter):
    items = [
        formatter.lfchar + formatter.htchar * (indent + 1) + formatter.format(item, indent + 1)
        for item in value
    ]
    return '(%s,)' % (','.join(items) + formatter.lfchar + formatter.htchar * indent)


class GenericFormatter(BaseFormatter):
    def can_format(self, value):
        return True

    def store(self, formatter, value):
        return GenericRepr.from_value(value)

    def format(self, value, indent, formatter):
        # `value` will always be a GenericRepr object because that's what `store` returns.
        return repr(value)

    def get_imports(self):
        return [('pysnap', 'GenericRepr')]

    def assert_value_matches_snapshot(self, test, test_value, snapshot_value):
        test_value = GenericRepr.from_value(test_value)
        # Assert equality between the representations to provide a nice textual diff.
        test.assert_equals(test_value.representation, snapshot_value.representation)


def default_formatters():
    return [
        TypeFormatter(type(None), format_none),
        TypeFormatter(dict, format_dict),
        TypeFormatter(tuple, format_tuple),
        TypeFormatter(list, format_list),
        TypeFormatter(six.string_types, format_str),
        TypeFormatter((int, float, complex, bool, bytes, set, frozenset), format_std_type),
        GenericFormatter()
    ]



import os
import shutil
import filecmp

from .formatter import Formatter
from .formatters import BaseFormatter


class FileSnapshot(object):
    def __init__(self, path):
        """
        Create a file snapshot pointing to the specified `path`. In a snapshot, `path` is considered to be relative to
        the test module's "snapshots" folder. (This is done to prevent ugly path manipulations inside the snapshot
        file.)
        """
        self.path = path

    def __repr__(self):
        return 'FileSnapshot({})'.format(repr(self.path))

    def __eq__(self, other):
        return self.path == other.path


class FileSnapshotFormatter(BaseFormatter):
    def can_format(self, value):
        return isinstance(value, FileSnapshot)

    def store(self, test, value):
        """
        Copy the file from the test location to the snapshot location.
        If the original test file has an extension, the snapshot file will use the same extension.
        """

        file_snapshot_dir = self.get_file_snapshot_dir(test)
        if not os.path.exists(file_snapshot_dir):
            os.makedirs(file_snapshot_dir, 0o0700)
        extension = os.path.splitext(value.path)[1]
        snapshot_file = os.path.join(file_snapshot_dir, test.test_name) + extension
        shutil.copy(value.path, snapshot_file)
        relative_snapshot_filename = os.path.relpath(snapshot_file, test.module.snapshot_dir)
        return FileSnapshot(relative_snapshot_filename)

    def get_imports(self):
        return (('pysnap.file', 'FileSnapshot'),)

    def format(self, value, indent, formatter):
        return repr(value)

    def assert_value_matches_snapshot(self, test, test_value, snapshot_value):
        snapshot_path = os.path.join(test.module.snapshot_dir, snapshot_value.path)
        files_identical = filecmp.cmp(test_value.path, snapshot_path, shallow=False)
        assert files_identical, "Stored file differs from test file"

    @staticmethod
    def get_file_snapshot_dir(test):
        """
        Get the directory for storing file snapshots for `test`.
        Snapshot files are stored under:
            snapshots/snap_<test_module_name>/
        Right next to where the snapshot module is stored:
            snapshots/snap_<snapshot_module_name>.py
        """
        return os.path.join(test.module.snapshot_dir, test.module.module)


Formatter.register_formatter(FileSnapshotFormatter())



class GenericRepr(object):
    def __init__(self, representation):
        self.representation = representation

    def __repr__(self):
        return 'GenericRepr({})'.format(repr(self.representation))

    def __eq__(self, other):
        return isinstance(other, GenericRepr) and self.representation == other.representation

    @staticmethod
    def from_value(value):
        representation = repr(value)
        # Remove the hex id, if found.
        representation = representation.replace(hex(id(value)), "0x100000000")
        return GenericRepr(representation)



from __future__ import unicode_literals


class SnapshotError(Exception):
    pass


class SnapshotNotFound(SnapshotError):
    def __init__(self, module, test_name):
        super(SnapshotNotFound, self).__init__(
            "Snapshot '{snapshot_id!s}' not found in {snapshot_file!s}".format(
                snapshot_id=test_name, snapshot_file=module.filepath))



from .formatters import default_formatters


class Formatter(object):
    formatters = default_formatters()

    def __init__(self, imports=None):
        self.htchar = ' '*4
        self.lfchar = '\n'
        self.indent = 0
        self.imports = imports

    def __call__(self, value, **args):
        return self.format(value, self.indent)

    def format(self, value, indent):
        formatter = self.get_formatter(value)
        for module, import_name in formatter.get_imports():
            self.imports[module].add(import_name)
        return formatter.format(value, indent, self)

    @staticmethod
    def get_formatter(value):
        for formatter in Formatter.formatters:
            if formatter.can_format(value):
                return formatter

        # This should never happen as GenericFormatter is registered by default.
        raise RuntimeError("No formatter found for value")

    @staticmethod
    def register_formatter(formatter):
        Formatter.formatters.insert(0, formatter)
